@@ -3,15 +3,17 @@
 
 use crate::{
     cargo_cli::{CargoCli, CargoOptions},
-    output::{OutputContext, OutputOpts},
+    output::{OutputContext, OutputOpts, OutputWriter},
+    reuse_build::ReuseBuildOpts,
     ExpectedError,
 };
 use camino::{Utf8Path, Utf8PathBuf};
 use clap::{ArgEnum, Args, Parser, Subcommand};
 use color_eyre::eyre::{Report, Result, WrapErr};
 use guppy::graph::PackageGraph;
+use nextest_metadata::{BinaryListSummary, BuildPlatform};
 use nextest_runner::{
-    config::NextestConfig,
+    config::{NextestConfig, NextestProfile},
     errors::{TargetRunnerError, WriteEventError},
     partition::PartitionerBuilder,
     reporter::{StatusLevel, TestOutputDisplay, TestReporterBuilder},
@@ -19,13 +21,13 @@ use nextest_runner::{
     signal::SignalHandler,
     target_runner::TargetRunner,
     test_filter::{RunIgnored, TestFilterBuilder},
-    test_list::{OutputFormat, RustTestArtifact, SerializableFormat, TestList},
+    test_list::{BinaryList, OutputFormat, RustTestArtifact, SerializableFormat, TestList},
 };
 use owo_colors::{OwoColorize, Style};
 use std::{
     error::Error,
     fmt::Write as _,
-    io::{BufWriter, Cursor, Write},
+    io::{Cursor, Write},
 };
 use supports_color::Stream;
 
@@ -42,21 +44,21 @@ pub struct CargoNextestApp {
 
 impl CargoNextestApp {
     /// Executes the app.
-    pub fn exec(self) -> Result<()> {
+    pub fn exec(self, output_writer: &mut OutputWriter) -> Result<()> {
         let NextestSubcommand::Nextest(app) = self.subcommand;
-        app.exec()
+        app.exec(output_writer)
     }
 }
 
 #[derive(Debug, Subcommand)]
 enum NextestSubcommand {
     /// A next-generation test runner for Rust. <https://nexte.st>
-    Nextest(AppImpl),
+    Nextest(AppOpts),
 }
 
 #[derive(Debug, Args)]
 #[clap(version)]
-struct AppImpl {
+struct AppOpts {
     /// Path to Cargo.toml
     #[clap(long, global = true, value_name = "PATH")]
     manifest_path: Option<Utf8PathBuf>,
@@ -71,6 +73,52 @@ struct AppImpl {
     command: Command,
 }
 
+impl AppOpts {
+    /// Execute the command.
+    fn exec(self, output_writer: &mut OutputWriter) -> Result<()> {
+        match self.command {
+            Command::List {
+                build_filter,
+                message_format,
+                list_type,
+                reuse_build,
+            } => {
+                let app = App::new(
+                    self.output,
+                    reuse_build,
+                    build_filter,
+                    self.config_opts,
+                    self.manifest_path,
+                )?;
+                app.exec_list(message_format, list_type, output_writer)
+            }
+            Command::Run {
+                profile,
+                no_capture,
+                build_filter,
+                runner_opts,
+                reporter_opts,
+                reuse_build,
+            } => {
+                let app = App::new(
+                    self.output,
+                    reuse_build,
+                    build_filter,
+                    self.config_opts,
+                    self.manifest_path,
+                )?;
+                app.exec_run(
+                    profile.as_deref(),
+                    no_capture,
+                    &runner_opts,
+                    &reporter_opts,
+                    output_writer,
+                )
+            }
+        }
+    }
+}
+
 #[derive(Debug, Args)]
 struct ConfigOpts {
     /// Config file [default: workspace-root/.config/nextest.toml]
@@ -108,6 +156,19 @@ enum Command {
             value_name = "FMT"
         )]
         message_format: MessageFormatOpts,
+
+        /// Type of listing
+        #[clap(
+            long,
+            arg_enum,
+            default_value_t,
+            help_heading = "OUTPUT OPTIONS",
+            value_name = "TYPE"
+        )]
+        list_type: ListType,
+
+        #[clap(flatten)]
+        reuse_build: ReuseBuildOpts,
     },
     /// Build and run tests
     ///
@@ -137,9 +198,47 @@ enum Command {
 
         #[clap(flatten)]
         reporter_opts: TestReporterOpts,
+
+        #[clap(flatten)]
+        reuse_build: ReuseBuildOpts,
     },
 }
 
+#[derive(Copy, Clone, Debug, ArgEnum)]
+pub(crate) enum PlatformFilterOpts {
+    Target,
+    Host,
+    Any,
+}
+
+impl Default for PlatformFilterOpts {
+    fn default() -> Self {
+        Self::Any
+    }
+}
+
+impl From<PlatformFilterOpts> for Option<BuildPlatform> {
+    fn from(opt: PlatformFilterOpts) -> Self {
+        match opt {
+            PlatformFilterOpts::Target => Some(BuildPlatform::Target),
+            PlatformFilterOpts::Host => Some(BuildPlatform::Host),
+            PlatformFilterOpts::Any => None,
+        }
+    }
+}
+
+#[derive(Copy, Clone, Debug, ArgEnum)]
+enum ListType {
+    Full,
+    BinariesOnly,
+}
+
+impl Default for ListType {
+    fn default() -> Self {
+        Self::Full
+    }
+}
+
 #[derive(Copy, Clone, Debug, ArgEnum)]
 enum MessageFormatOpts {
     Human,
@@ -177,6 +276,10 @@ struct TestBuildFilter {
     #[clap(long)]
     partition: Option<PartitionerBuilder>,
 
+    /// Filter test binaries by build platform
+    #[clap(long, arg_enum, value_name = "PLATFORM", default_value_t)]
+    pub(crate) platform_filter: PlatformFilterOpts,
+
     // TODO: add regex-based filtering in the future?
     /// Test name filter
     #[clap(name = "FILTERS", help_heading = None)]
@@ -184,13 +287,31 @@ struct TestBuildFilter {
 }
 
 impl TestBuildFilter {
-    fn compute<'g>(
+    fn compute_test_list<'g>(
         &self,
-        manifest_path: Option<&'g Utf8Path>,
         graph: &'g PackageGraph,
-        output: OutputContext,
-        runner: Option<&TargetRunner>,
+        binary_list: BinaryList,
+        runner: &TargetRunner,
+        reuse_build: &ReuseBuildOpts,
     ) -> Result<TestList<'g>> {
+        let path_mapper = reuse_build.make_path_mapper(graph);
+        let test_artifacts = RustTestArtifact::from_binary_list(
+            graph,
+            binary_list,
+            path_mapper.as_ref(),
+            self.platform_filter.into(),
+        )?;
+        let test_filter =
+            TestFilterBuilder::new(self.run_ignored, self.partition.clone(), &self.filter);
+        TestList::new(test_artifacts, &test_filter, runner).wrap_err("error building test list")
+    }
+
+    fn compute_binary_list(
+        &self,
+        graph: &PackageGraph,
+        manifest_path: Option<&Utf8Path>,
+        output: OutputContext,
+    ) -> Result<BinaryList> {
         // Don't use the manifest path from the graph to ensure that if the user cd's into a
         // particular crate and runs cargo nextest, then it behaves identically to cargo test.
         let mut cargo_cli = CargoCli::new("test", manifest_path, output);
@@ -212,11 +333,8 @@ impl TestBuildFilter {
             )));
         }
 
-        let test_artifacts = RustTestArtifact::from_messages(graph, Cursor::new(output.stdout))?;
-
-        let test_filter =
-            TestFilterBuilder::new(self.run_ignored, self.partition.clone(), &self.filter);
-        TestList::new(test_artifacts, &test_filter, runner).wrap_err("error building test list")
+        let test_binaries = BinaryList::from_messages(Cursor::new(output.stdout), graph)?;
+        Ok(test_binaries)
     }
 }
 
@@ -311,92 +429,182 @@ impl TestReporterOpts {
     }
 }
 
-impl AppImpl {
-    /// Execute the command.
-    fn exec(self) -> Result<()> {
-        let output = self.output.init();
+struct App {
+    output: OutputContext,
+    graph: PackageGraph,
+    workspace_root: Utf8PathBuf,
+    manifest_path: Option<Utf8PathBuf>,
+    reuse_build: ReuseBuildOpts,
+    build_filter: TestBuildFilter,
+    config_opts: ConfigOpts,
+}
 
-        let graph = build_graph(self.manifest_path.as_deref(), output)?;
+impl App {
+    fn new(
+        output: OutputOpts,
+        reuse_build: ReuseBuildOpts,
+        build_filter: TestBuildFilter,
+        config_opts: ConfigOpts,
+        manifest_path: Option<Utf8PathBuf>,
+    ) -> Result<Self> {
+        let output = output.init();
+        reuse_build.check_experimental(output)?;
 
-        match self.command {
-            Command::List {
-                build_filter,
-                message_format,
-            } => {
-                let target_runner = runner_for_target(build_filter.cargo_options.target.as_deref());
+        let graph_data = match reuse_build.cargo_metadata.as_deref() {
+            Some(path) => std::fs::read_to_string(path)?,
+            None => acquire_graph_data(manifest_path.as_deref(), output)?,
+        };
+        let graph = guppy::CargoMetadata::parse_json(&graph_data)?.build_graph()?;
 
-                let mut test_list = build_filter.compute(
-                    self.manifest_path.as_deref(),
-                    &graph,
-                    output,
-                    target_runner.as_ref(),
+        let manifest_path = if reuse_build.cargo_metadata.is_some() {
+            Some(graph.workspace().root().join("Cargo.toml"))
+        } else {
+            manifest_path
+        };
+
+        let workspace_root = match &reuse_build.workspace_remap {
+            Some(path) => path.clone(),
+            _ => graph.workspace().root().to_owned(),
+        };
+
+        Ok(Self {
+            output,
+            graph,
+            reuse_build,
+            build_filter,
+            manifest_path,
+            workspace_root,
+            config_opts,
+        })
+    }
+
+    fn build_binary_list(&self) -> Result<BinaryList> {
+        let binary_list = match self.reuse_build.binaries_metadata.as_deref() {
+            Some(path) => {
+                let raw_binary_list = std::fs::read_to_string(path)?;
+                let binary_list: BinaryListSummary = serde_json::from_str(&raw_binary_list)?;
+                BinaryList::from_summary(binary_list)
+            }
+            None => self.build_filter.compute_binary_list(
+                &self.graph,
+                self.manifest_path.as_deref(),
+                self.output,
+            )?,
+        };
+        Ok(binary_list)
+    }
+
+    fn build_test_list(
+        &self,
+        binary_list: BinaryList,
+        target_runner: &TargetRunner,
+    ) -> Result<TestList> {
+        self.build_filter.compute_test_list(
+            &self.graph,
+            binary_list,
+            target_runner,
+            &self.reuse_build,
+        )
+    }
+
+    fn load_profile<'cfg>(
+        &self,
+        profile_name: Option<&str>,
+        config: &'cfg NextestConfig,
+    ) -> Result<NextestProfile<'cfg>> {
+        let profile = config
+            .profile(profile_name.unwrap_or(NextestConfig::DEFAULT_PROFILE))
+            .map_err(ExpectedError::profile_not_found)?;
+        let store_dir = profile.store_dir();
+        std::fs::create_dir_all(&store_dir)
+            .wrap_err_with(|| format!("failed to create store dir '{}'", store_dir))?;
+        Ok(profile)
+    }
+
+    fn load_runner(&self) -> TargetRunner {
+        // When cross-compiling we should not use the cross target runner
+        // for running the host tests (like proc-macro ones).
+        runner_for_target(self.build_filter.cargo_options.target.as_deref())
+    }
+
+    fn exec_list(
+        &self,
+        message_format: MessageFormatOpts,
+        list_type: ListType,
+        output_writer: &mut OutputWriter,
+    ) -> Result<()> {
+        let binary_list = self.build_binary_list()?;
+
+        match list_type {
+            ListType::BinariesOnly => {
+                let mut writer = output_writer.stdout_writer();
+                binary_list.write(
+                    message_format.to_output_format(self.output.verbose),
+                    &mut writer,
+                    self.output.color.should_colorize(Stream::Stdout),
                 )?;
-                if output.color.should_colorize(Stream::Stdout) {
-                    test_list.colorize();
-                }
-                let stdout = std::io::stdout();
-                let lock = stdout.lock();
-                // Buffer the output to minimize syscalls.
-                let mut writer = BufWriter::new(lock);
-                test_list.write(message_format.to_output_format(output.verbose), &mut writer)?;
                 writer.flush()?;
             }
-            Command::Run {
-                ref profile,
-                no_capture,
-                ref build_filter,
-                ref runner_opts,
-                ref reporter_opts,
-            } => {
-                let config = self.config_opts.make_config(graph.workspace().root())?;
-                let profile = config
-                    .profile(profile.as_deref().unwrap_or(NextestConfig::DEFAULT_PROFILE))
-                    .map_err(ExpectedError::profile_not_found)?;
-                let store_dir = profile.store_dir();
-                std::fs::create_dir_all(&store_dir)
-                    .wrap_err_with(|| format!("failed to create store dir '{}'", store_dir))?;
+            ListType::Full => {
+                let target_runner = self.load_runner();
+                let test_list = self.build_test_list(binary_list, &target_runner)?;
 
-                let target_runner = runner_for_target(build_filter.cargo_options.target.as_deref());
-
-                let test_list = build_filter.compute(
-                    self.manifest_path.as_deref(),
-                    &graph,
-                    output,
-                    target_runner.as_ref(),
+                let mut writer = output_writer.stdout_writer();
+                test_list.write(
+                    message_format.to_output_format(self.output.verbose),
+                    &mut writer,
+                    self.output.color.should_colorize(Stream::Stdout),
                 )?;
-
-                let mut reporter = reporter_opts
-                    .to_builder(no_capture)
-                    .set_verbose(output.verbose)
-                    .build(&test_list, &profile);
-                if output.color.should_colorize(Stream::Stderr) {
-                    reporter.colorize();
-                }
-
-                let handler = SignalHandler::new().wrap_err("failed to set up Ctrl-C handler")?;
-                let mut runner_builder = runner_opts.to_builder(no_capture);
-                if let Some(target_runner) = target_runner {
-                    runner_builder.set_target_runner(target_runner);
-                }
-
-                let runner = runner_builder.build(&test_list, &profile, handler);
-                let stderr = std::io::stderr();
-                let mut writer = BufWriter::new(stderr);
-                let run_stats = runner.try_execute(|event| {
-                    // Write and flush the event.
-                    reporter.report_event(event, &mut writer)?;
-                    writer.flush().map_err(WriteEventError::Io)
-                })?;
-                if !run_stats.is_success() {
-                    return Err(Report::new(ExpectedError::test_run_failed()));
-                }
+                writer.flush()?;
             }
         }
         Ok(())
     }
+
+    fn exec_run(
+        &self,
+        profile_name: Option<&str>,
+        no_capture: bool,
+        runner_opts: &TestRunnerOpts,
+        reporter_opts: &TestReporterOpts,
+        output_writer: &mut OutputWriter,
+    ) -> Result<()> {
+        let config = self
+            .config_opts
+            .make_config(self.workspace_root.as_path())?;
+        let profile = self.load_profile(profile_name, &config)?;
+
+        let target_runner = self.load_runner();
+
+        let binary_list = self.build_binary_list()?;
+        let test_list = self.build_test_list(binary_list, &target_runner)?;
+
+        let mut reporter = reporter_opts
+            .to_builder(no_capture)
+            .set_verbose(self.output.verbose)
+            .build(&test_list, &profile);
+        if self.output.color.should_colorize(Stream::Stderr) {
+            reporter.colorize();
+        }
+
+        let handler = SignalHandler::new().wrap_err("failed to set up Ctrl-C handler")?;
+        let runner_builder = runner_opts.to_builder(no_capture);
+        let runner = runner_builder.build(&test_list, &profile, handler, target_runner);
+
+        let mut writer = output_writer.stderr_writer();
+        let run_stats = runner.try_execute(|event| {
+            // Write and flush the event.
+            reporter.report_event(event, &mut writer)?;
+            writer.flush().map_err(WriteEventError::Io)
+        })?;
+        if !run_stats.is_success() {
+            return Err(Report::new(ExpectedError::test_run_failed()));
+        }
+        Ok(())
+    }
 }
 
-fn build_graph(manifest_path: Option<&Utf8Path>, output: OutputContext) -> Result<PackageGraph> {
+fn acquire_graph_data(manifest_path: Option<&Utf8Path>, output: OutputContext) -> Result<String> {
     let mut cargo_cli = CargoCli::new("metadata", manifest_path, output);
     // Construct a package graph with --no-deps since we don't need full dependency
     // information.
@@ -415,15 +623,15 @@ fn build_graph(manifest_path: Option<&Utf8Path>, output: OutputContext) -> Resul
 
     let json =
         String::from_utf8(output.stdout).wrap_err("cargo metadata output is invalid UTF-8")?;
-    Ok(guppy::CargoMetadata::parse_json(&json)?.build_graph()?)
+    Ok(json)
 }
 
-fn runner_for_target(triple: Option<&str>) -> Option<TargetRunner> {
-    match TargetRunner::for_target(triple) {
+fn runner_for_target(triple: Option<&str>) -> TargetRunner {
+    match TargetRunner::new(triple) {
         Ok(runner) => runner,
         Err(err) => {
             warn_on_target_runner_err(&err).expect("writing to a string is infallible");
-            None
+            TargetRunner::empty()
         }
     }
 }
